﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Xml.Linq;

using static System.IO.File;

namespace UMK
{
    public class SettingsManager<T> where T : new()
    {
        private struct ObjDescription
        {
            public ObjDescription(Type type, Dictionary<string, object> values)
            {
                this.Type = type;
                this.Values = values;
            }

            public Type Type { get; private set; }
            public Dictionary<string, object> Values { get; private set; }
        }

        //private T settingsObject;

        private Dictionary<string, object> values = new Dictionary<string, object>();
        private string path;
        private bool useDefaultValuesInsteadOfThrowingException;
        private IFormatProvider formatProvider;

        private SettingsManager(Dictionary<string, object> values, string path, bool useDefaultValuesInsteadOfThrowingException = false, IFormatProvider formatProvider = null)
        {
            if (formatProvider == null) 
                formatProvider = System.Globalization.CultureInfo.InvariantCulture;

            this.values = values;
            this.path = path;
            this.useDefaultValuesInsteadOfThrowingException = useDefaultValuesInsteadOfThrowingException;
            this.formatProvider = formatProvider;
        }

        public SettingsManager(T settingsObject, string path, bool useDefaultValuesInsteadOfThrowingException = false, IFormatProvider formatProvider = null)
            : this(ExtractValues(settingsObject), path, useDefaultValuesInsteadOfThrowingException, formatProvider)
        {
            //tu ma nic nie być
        }

        private static bool IsNeastedType(Type type)
        {
            bool result = FieldsAndProperties(type).Any(m => m.IsProper) && !type.IsEnum;
            return result;
        }

        private static IEnumerable<IMemberAdapter> FieldsAndProperties(Type type)
        {
            return type.GetFields().Select(f => (IMemberAdapter)new FieldAdapter(f))
                .Concat(type.GetProperties().Select(p => new PropertyAdapter(p)));
        }

        //Reflection
        public static Dictionary<string, object> ExtractValues(object settingsObject)
        {
            if (settingsObject == null) 
                throw new SettingsException("Empty object");

            Dictionary<string, object> values = new Dictionary<string, object>();

            //FieldInfo[] fields = settingsObject.GetType().GetFields();            
            //FieldInfo[] fields = typeof(T).GetFields(BindingFlags.NonPublic | BindingFlags.Instance);
            IEnumerable<IMemberAdapter> fap = FieldsAndProperties(settingsObject.GetType());
            
            foreach (IMemberAdapter member in fap)
            {
                if (member.IsProper)
                {
                    object value = member.GetValue(settingsObject);

                    if (IsNeastedType(member.Type))
                        values.Add(member.Name, new ObjDescription(value.GetType(), ExtractValues(value)));
                    else
                        values.Add(member.Name, value);
                }
            }

            return values;
        }

        private static string ValueToString(object value, IFormatProvider formatProvider)
        {
            if (value.GetType().IsArray)
            {
                string s = "";
                Array array = value as Array;
                //foreach (object element in array) s += element.ToString() + ArrayElementsSeparator;
                foreach (object element in array) 
                    s += Convert.ToString(element, formatProvider) + ArrayElementsSeparator;
                s = s.Substring(0, s.Length - 1);
                return s;
            }
            else return Convert.ToString(value, formatProvider);
        }

        private XElement SerializeValues(Dictionary<string, object> valuesParam)
        {
            return new XElement("Settings", valuesParam.OrderBy(x => x.Key).Select(kvp => 
            {
                var setting = new XElement("Setting", new XAttribute("Name", kvp.Key));

                if (kvp.Value is ObjDescription neastedObj)
                {
                    setting.Add(new XAttribute("Type", neastedObj.Type.Name));
                    setting.Add(SerializeValues(neastedObj.Values));
                }
                else
                {
                    setting.Add(new XAttribute("Type", kvp.Value.GetType().Name));
                    setting.Add(ValueToString(kvp.Value, formatProvider));
                }

                return setting;
            }));
        }

        public void Save()
        {
            try
            {
                XDocument xml = new XDocument(
                    new XDeclaration("1.0", "utf-8", "yes"),
                    new XComment("utworzono " + DateTime.Now.ToString()),
                    SerializeValues(values)
                );

                xml.Save(path);
            }
            catch (Exception exc)
            {
                throw new SettingsException("Błąd przy zapisie do pliku XML", exc);
            }
        }

        public static SettingsManager<T> Load(string path, bool useDefaultValuesInsteadOfThrowingException = false, IFormatProvider formatProvider = null)
        {
            if (!Exists(path)) 
                throw new SettingsException("Brak pliku XML z ustawieniami");

            try
            {
                Dictionary<string, object> values = new Dictionary<string, object>();
                foreach (XElement element in XDocument.Load(path).Element("Settings").Elements("Setting"))
                {
                    string name = element.Attribute("Name").Value;
                    string value = element.HasElements ? element.FirstNode.ToString() : element.Value;
                    values.Add(name, value); //TODO: tu chyba powinna być konwersja na typ
                }
                return new SettingsManager<T>(values, path, useDefaultValuesInsteadOfThrowingException, formatProvider);
            }
            catch (Exception exc)
            {
                throw new SettingsException("Błąd przy odczycie pliku XML", exc);
            }
        }

        private static object BuildMemberObject(Type memberType, object memberValue, IFormatProvider formatProvider)
        {
            if (memberValue == null) 
                return null;

            string svalue = memberValue.ToString();
            if (memberType != typeof(string) && svalue.Equals(memberType.ToString()))
                throw new SettingsException("Podejrzewam, że wartość ma tylko nazwę typu");

            object value = null;
            if (memberType.IsEnum)
                value = Enum.Parse(memberType, svalue);
            else if (memberType.IsArray)
                value = BuildMemberArray(memberType, svalue, formatProvider);
            else if (IsNeastedType(memberType))
                value = BuildNeastedType(memberType, svalue);
            else
                value = Convert.ChangeType(memberValue, memberType, formatProvider);

            return value;
        }

        public static char ArrayElementsSeparator = ';';

        private static object BuildMemberArray(Type memberType, string svalue, IFormatProvider formatProvider)
        {
            string[] svalues = svalue.Split(ArrayElementsSeparator);
            Type elementType = memberType.GetElementType();
            Array instance = Array.CreateInstance(elementType, svalues.Length);
            for(int i = instance.GetLowerBound(0); i <= instance.GetUpperBound(0); ++i)
            {
                object o = Convert.ChangeType(svalues[i - instance.GetLowerBound(0)], elementType, formatProvider);
                instance.SetValue(o, i);
            }
            return instance;
        }

        private static object BuildNeastedType(Type type, string svalue)
        {
            Guid guid = Guid.NewGuid();
            string tempFileName = $"{guid}.xml";

            File.WriteAllText(tempFileName, svalue.ToString());
            Type targetSettingsManagerType = typeof(SettingsManager<>).MakeGenericType(type);

            var kopia = targetSettingsManagerType.GetMethod(nameof(Load)).Invoke(null, new object[] { tempFileName, false, null });
            object member = targetSettingsManagerType.GetMethod(nameof(GetSettingsObject)).Invoke(kopia, null);

            File.Delete(tempFileName);

            return member;
        }


        private static T BuildObject(Dictionary<string, object> values, bool useDefaultValuesInsteadOfThrowingException, IFormatProvider formatProvider)
        {
            object settingsObject = new T();

            IEnumerable<IMemberAdapter> members = FieldsAndProperties(typeof(T));
            foreach(IMemberAdapter member in members)
            {
                if(member.IsProper)
                {
                    object value = null;
                    if(values.ContainsKey(member.Name))
                    {
                        value = BuildMemberObject(member.Type, values[member.Name], formatProvider);
                        member.SetValue(settingsObject, value);
                    }
                }
            }

            return (T)settingsObject;
        }

        public T GetSettingsObject()
        {
            return BuildObject(values, useDefaultValuesInsteadOfThrowingException, formatProvider);
        }
    }    


    public class SettingsException : Exception
    {
        public SettingsException(string message = null, Exception innerException = null)
            : base(message, innerException)
        {
        }
    }
}
